from __future__ import annotations

from ninja import NinjaAPI, Schema

from allianceauth.framework.api.user import (
    get_all_characters_from_user, get_main_character_from_user,
)
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema


class WhoamiResponse(Schema):
    main_character_id: int
    access: list[str]
    characters: list[CharacterSchema]


logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    AuthAPIEndpoints(api)


class AuthAPIEndpoints:

    tags = ["Auth"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/auth/whoami", response={200: WhoamiResponse, 404: dict}, tags=self.tags)
        def whoami(request) -> tuple[int, WhoamiResponse | dict]:
            user = request.user
            logger.debug(f"Authenticating user: {user}")

            try:
                main_character = get_main_character_from_user(user)
            except Exception as e:
                logger.warning("Main character fetch failed", exc_info=e)
                return 404, {"error": "User has no Main Character Set"}

            access_levels: list[str] = [perm for perm in user.get_all_permissions() if perm.startswith("incursions.")]
            characters = get_all_characters_from_user(user)
            character_schemas = [CharacterSchema.from_orm(alt) for alt in characters]

            logger.info(f"User {user} authenticated with {len(character_schemas)} characters")
            return 200, WhoamiResponse(main_character_id=main_character.character_id, access=access_levels, characters=character_schemas)



# Generated by Django 4.2 on 2025-04-08 05:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0024_remove_incursion_constellation_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='incursion',
            old_name='migration_constellation',
            new_name='constellation',
        ),
        migrations.RenameField(
            model_name='incursion',
            old_name='migration_infested_solar_systems',
            new_name='infested_solar_systems',
        ),
        migrations.RenameField(
            model_name='incursion',
            old_name='migration_staging_solar_system',
            new_name='staging_solar_system',
        ),
    ]



# Generated by Django 4.2 on 2025-04-05 11:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0118_alter_skill_id'),
        ('incursions', '0019_incursion_migration_constellation_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='incursion',
            name='migration_infested_solar_systems',
            field=models.ManyToManyField(blank=True, related_name='+', to='corptools.mapsystem'),
        ),
    ]



from datetime import datetime, timezone

from celery import shared_task
from corptools.models import EveItemType, MapConstellation, MapSystem
from ninja import Schema

from django.db import IntegrityError, transaction
from django.db.models import Count
from django.utils.timezone import now

from allianceauth.eveonline.models import EveFactionInfo
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from incursions.api.helpers.sse import SSEClient, SSEEvent
from incursions.api.waitlist import WaitlistUpdateSchema
from incursions.app_settings import SSE_SECRET, SSE_SITE_URL
from incursions.helpers import (
    embed_boss_spawned, embed_ended, embed_established,
    embed_established_addendum, embed_mobilizing, embed_waitlist_state,
    embed_withdrawing,
)
from incursions.models.incursion import (
    Incursion, IncursionInfluence, IncursionsConfig,
)
from incursions.models.waitlist import (
    ActiveFleet, Fleet, FleetActivity, Waitlist, WaitlistEntry,
    WaitlistEntryFit,
)
from incursions.providers import get_incursions_incursions

logger = get_extension_logger(__name__)
sse_client = SSEClient(SSE_SITE_URL, SSE_SECRET)


class NewFleetCompSchema(Schema):
    fleet_id: int


@shared_task(base=QueueOnce)
def update_incursions() -> None:
    incursions, response = get_incursions_incursions()
    actives = []
    for incursion in incursions:
        actives.append(incursion['constellation_id'])
        try:
            # Get, because i need to do more steps than an update_or_create would let me
            # This chunk is purely for when incursions change states.
            # Also incursions have no unique id.... wtf ccp
            i = Incursion.objects.get(
                constellation=MapConstellation.objects.get(constellation_id=incursion['constellation_id']),
                ended_timestamp__isnull=True)
            if incursion['state'] == "established":
                # This is still just an established incursion, nothing to act on
                pass
            elif incursion['state'] == "mobilizing" and i.state != Incursion.States.MOBILIZING:
                i.mobilizing_timestamp = datetime.strptime(
                    str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc)
                i.state = Incursion.States.MOBILIZING
                if i.has_boss == "true":
                    i.has_boss = True
                i.save(update_fields=["mobilizing_timestamp", "state"])
            elif incursion['state'] == "withdrawing" and i.state != Incursion.States.WITHDRAWING:
                i.withdrawing_timestamp = datetime.strptime(
                    str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc)
                i.state = Incursion.States.WITHDRAWING
                if i.has_boss == "true":
                    i.has_boss = True
                i.save(update_fields=["withdrawing_timestamp", "state"])
            else:
                # ????
                pass
            try:
                IncursionInfluence.objects.create(
                    incursion=i,
                    influence=incursion['influence'],
                    timestamp=datetime.strptime(
                        str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc))
            except IntegrityError:
                # If we call this task too often cache will return the same influence
                pass
        except Incursion.DoesNotExist:
            # Create an Incursion, It does not exist.
            i = Incursion.objects.create(
                constellation=MapConstellation.objects.get(constellation_id=incursion['constellation_id']),
                faction=EveFactionInfo.objects.get_or_create(faction_id=incursion['faction_id'])[0],
                has_boss=True if incursion['has_boss'] == "true" else False,
                staging_solar_system=MapSystem.objects.get(system_id=incursion['staging_solar_system_id'])[0],
                state=incursion['state'],
                type=incursion['type'],
                established_timestamp=datetime.strptime(
                    str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc)
            )
            # We need to also set the mobilizing and withdrawing state here
            # This is purely for new installs, bcoz partially complete incursions
            if incursion['state'] == "mobilizing":
                i.mobilizing_timestamp = datetime.strptime(
                    str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc)
                i.save(update_fields=["mobilizing_timestamp"])
            elif incursion['state'] == "withdrawing ":
                i.mobilizing_timestamp = datetime.strptime(
                    str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc)
                i.withdrawing_timestamp = datetime.strptime(
                    str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc)
                i.save(update_fields=["withdrawing_timestamp", "mobilizing_timestamp"])
            try:
                IncursionInfluence.objects.create(
                    incursion=i,
                    influence=incursion['influence'],
                    timestamp=datetime.strptime(
                        str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z').replace(tzinfo=timezone.utc))
            except IntegrityError:
                # If we call this task too often cache will return the same influence
                pass

    for ended in Incursion.objects.filter(ended_timestamp__isnull=True).exclude(constellation_id__in=actives):
        # Cant use update here, need to fire signals
        ended.ended_timestamp = datetime.strptime(
            str(response.headers['Last-Modified']), '%a, %d %b %Y %H:%M:%S %Z')
        ended.state = Incursion.States.ENDED
        ended.save(update_fields=["ended_timestamp", "state"])


@shared_task
def incursion_established(incursion_pk: int) -> None:
    incursion = Incursion.objects.get(pk=incursion_pk)
    for webhook in IncursionsConfig.get_solo().status_webhooks.all():
        if incursion.staging_solar_system.is_high_sec == webhook.security_high:
            webhook.send_embed(embed=embed_established(incursion))
            webhook.send_embed(embed=embed_established_addendum(incursion))
        elif incursion.staging_solar_system.is_low_sec == webhook.security_low:
            webhook.send_embed(embed=embed_established(incursion))
            webhook.send_embed(embed=embed_established_addendum(incursion))
        elif incursion.staging_solar_system.is_null_sec == webhook.security_null:
            webhook.send_embed(embed=embed_established(incursion))
            webhook.send_embed(embed=embed_established_addendum(incursion))


@shared_task
def incursion_mobilizing(incursion_pk: int) -> None:
    incursion = Incursion.objects.get(pk=incursion_pk)
    for webhook in IncursionsConfig.get_solo().status_webhooks.all():
        if incursion.staging_solar_system.is_high_sec == webhook.security_high:
            webhook.send_embed(embed=embed_mobilizing(incursion))
        elif incursion.staging_solar_system.is_low_sec == webhook.security_low:
            webhook.send_embed(embed=embed_mobilizing(incursion))
        elif incursion.staging_solar_system.is_null_sec == webhook.security_null:
            webhook.send_embed(embed=embed_mobilizing(incursion))


@shared_task
def incursion_withdrawing(incursion_pk: int) -> None:
    incursion = Incursion.objects.get(pk=incursion_pk)
    for webhook in IncursionsConfig.get_solo().status_webhooks.all():
        if incursion.staging_solar_system.is_high_sec == webhook.security_high:
            webhook.send_embed(embed=embed_withdrawing(incursion))
        elif incursion.staging_solar_system.is_low_sec == webhook.security_low:
            webhook.send_embed(embed=embed_withdrawing(incursion))
        elif incursion.staging_solar_system.is_null_sec == webhook.security_null:
            webhook.send_embed(embed=embed_withdrawing(incursion))


@shared_task
def incursion_ended(incursion_pk: int) -> None:
    incursion = Incursion.objects.get(pk=incursion_pk)
    for webhook in IncursionsConfig.get_solo().status_webhooks.all():
        if incursion.staging_solar_system.is_high_sec == webhook.security_high:
            webhook.send_embed(embed=embed_ended(incursion))
        elif incursion.staging_solar_system.is_low_sec == webhook.security_low:
            webhook.send_embed(embed=embed_ended(incursion))
        elif incursion.staging_solar_system.is_null_sec == webhook.security_null:
            webhook.send_embed(embed=embed_ended(incursion))


@shared_task
def incursion_boss_spawned(incursion_pk: int) -> None:
    incursion = Incursion.objects.get(pk=incursion_pk)
    for webhook in IncursionsConfig.get_solo().status_webhooks.all():
        if incursion.staging_solar_system.is_high_sec == webhook.security_high:
            webhook.send_embed(embed=embed_boss_spawned(incursion))
        elif incursion.staging_solar_system.is_low_sec == webhook.security_low:
            webhook.send_embed(embed=embed_boss_spawned(incursion))
        elif incursion.staging_solar_system.is_null_sec == webhook.security_null:
            webhook.send_embed(embed=embed_boss_spawned(incursion))


@shared_task
def waitlist_state(waitlist_pk: int, is_open: bool) -> None:
    waitlist = Waitlist.get_solo()
    for webhook in IncursionsConfig.get_solo().waitlist_webhooks.all():
        webhook.send_embed(embed=embed_waitlist_state(waitlist=waitlist, is_open=is_open))


@shared_task(base=QueueOnce)
def update_all_fleets() -> None:

    fleet = ActiveFleet.get_solo().fleet
    update_fleet(fleet.pk)


def update_fleet(fleet_id: int) -> None:

    fleet = Fleet.objects.select_related('boss').get(pk=fleet_id)

    if fleet.is_updating is False:
        return

    new_comp_detected = False
    changed_waitlists = False
    waitlist = Waitlist.get_solo()

    # ARIEL: this wants much more error handling of various cases
    # Sometimes ESI is unavailable etc, rn it fails once and shuts down
    try:
        esi_fleet = fleet.get_fleet_members()
        fleet.last_updated = now()
        fleet.save(update_fields=["last_updated"])
    except Exception as e:
        logger.error(f"Fleet {fleet.pk} failed to get fleet esi: {e}")
        fleet.is_updating = False
        fleet.save(update_fields=["is_updating"])
        return

    fleet_members = {m['character_id']: m for m in esi_fleet}
    fleet_member_ids = list(fleet_members.keys())

    old_fleet_members = FleetActivity.objects.filter(fleet=fleet, has_left=False).values_list('character_id', flat=True)

    # Clean waitlist
    with transaction.atomic():
        count, deleted = WaitlistEntryFit.objects.filter(character__character_id__in=fleet_member_ids).delete()
        WaitlistEntry.objects.annotate(fit_count=Count('waitlistentryfit')).filter(fit_count=0).delete()

    if count > 0:
        changed_waitlists = True

    # Update fleet composition
    with transaction.atomic():

        min_required = 5  # ARIEL replace this with a config

        if len(fleet_members) < min_required:
            return

        for char_id, fleet_member in fleet_members.items():
            fa, created = FleetActivity.objects.update_or_create(
                character_id=char_id,
                fleet=fleet,
                ship_id=fleet_member['ship_type_id'],
                defaults={
                    "ship": EveItemType.objects.get_or_create_from_esi(eve_id=fleet_member['ship_type_id'])[0],
                    # "first_seen": now(),  # auto_now_add
                    "last_seen": now(),
                    "is_boss": fleet.boss.character_id == char_id,
                    "has_left": False,
                })
            if created:
                new_comp_detected = True

        updated_count = FleetActivity.objects.filter(
            character_id__in=old_fleet_members,
            fleet=fleet
        ).exclude(
            character_id__in=fleet_member_ids
        ).update(
            has_left=True,
            last_seen=now(),
        )
        if updated_count > 0:
            new_comp_detected = True

    # Notify frontend via SSE or signals

    if new_comp_detected:
        sse_payload = NewFleetCompSchema(fleet_id=waitlist.pk).model_dump()
        sse_client.submit([SSEEvent.new_json(topic="fleet_comp", event="comp_update", data=sse_payload)])
    if changed_waitlists:
        sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
        sse_client.submit([SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)])



# Generated by Django 4.2 on 2025-04-09 05:00

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
        ('incursions', '0025_rename_migration_constellation_incursion_constellation_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='badge',
            name='color',
            field=models.CharField(choices=[('Red', 'Red'), ('Green', 'Green'), ('Blue', 'Blue'), ('Yellow', 'Yellow'), ('Purple', 'Purple'), ('Neutral', 'Neutral'), ('Cyan', 'Cyan')], default='Neutral', max_length=16, verbose_name='Badge Color'),
        ),
        migrations.AddField(
            model_name='badge',
            name='letter',
            field=models.CharField(default='?', max_length=1, verbose_name='Letter'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='badge',
            name='order',
            field=models.SmallIntegerField(default=0, help_text='Order in which the badge will be displayed, lower is earlier', verbose_name='Order'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='badge',
            name='type',
            field=models.CharField(choices=[('Shield', 'Shield')], default='Shield', max_length=16, verbose_name='Badge Type'),
        ),
        migrations.AlterField(
            model_name='characterbadges',
            name='character',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='incursions_badge', to='eveonline.evecharacter', verbose_name='Character'),
        ),
        migrations.DeleteModel(
            name='CharacterRoles',
        ),
        migrations.DeleteModel(
            name='Role',
        ),
    ]



from django.db import models


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_waitlist", "Waitlist Access"),  # All
            ("waitlist_alts_view", "View - all of a users Alts ! Danger?"),  # Leadership
            ("waitlist_announcements_manage", "Announcements - Manage"),  # Leadership
            ("waitlist_badges_view", "Badges - View"),  # FC
            ("waitlist_badges_manage", "Badges - Manage"),  # Leadership
            ("waitlist_bans_view", "Bans - View"),  # FC
            ("waitlist_bans_manage", "Bans - Manage"),  # Leadership
            ("waitlist_documentation_view", "Documentation - View"),  # Jr FC
            ("waitlist_fleet_view", "Fleet Composition"),  # FC
            ("waitlist_history_view", "Fleet History"),  # Leadership
            ("waitlist_notes_view", "Notes - View"),  # FC
            ("waitlist_notes_manage", "Notes - Manage"),  # Leadership
            ("waitlist_search", "Search"),  # FC
            ("waitlist_stats_view", "Fleet Stats"),  # Leadership
            ("waitlist_skills_view", "Skills"),  # Leadership
            ("waitlist_pilot_view", "Pilot Information"),  # Leadership
            # Add extra context to the Waitlist for Non-FCs
            ("waitlist_context_a", "Context: Number of Pilots"),
            ("waitlist_context_b", "Context: Ship Types"),
            ("waitlist_context_c", "Context: Time in Waitlist"),
            ("waitlist_context_d", "Context: Pilot Names"),
            ("waitlist_implants_view", "Context: Implants"),  # FC

            ("waitlist_manage_waitlist", "Waitlist Manage"),  # FC
            ("waitlist_manage_waitlist_approve_fits", "Waitlist Manage Fits"),  # FC

            # ESI Calls, Limit these maybe?
            ("waitlist_esi_show_info", "Show Info"),
            ("waitlist_esi_search", "Search - ESI"),
        )



# Generated by Django 4.2 on 2025-04-09 06:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0026_badge_color_badge_letter_badge_order_badge_type_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='waitlistentryfit',
            name='tags',
            field=models.CharField(help_text='Comma Seperated: Tags are fit specific Badges, such as First/New/Alt/Skills', max_length=255, verbose_name='Tags'),
        ),
        migrations.DeleteModel(
            name='SkillHistory',
        ),
    ]



# Generated by Django 4.2 on 2025-04-08 05:30

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0023_alter_approvedskills_options_alter_badge_options_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='incursion',
            name='constellation',
        ),
        migrations.RemoveField(
            model_name='incursion',
            name='infested_solar_systems',
        ),
        migrations.RemoveField(
            model_name='incursion',
            name='staging_solar_system',
        ),
        migrations.AlterField(
            model_name='fleet',
            name='last_updated',
            field=models.DateTimeField(blank=True, verbose_name='Last Updated'),
        ),
        migrations.AlterField(
            model_name='fleetactivity',
            name='last_seen',
            field=models.DateTimeField(verbose_name='Last Seen'),
        ),
        migrations.AlterField(
            model_name='waitlistentryfit',
            name='waitlist_entry',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incursions.waitlistentry', verbose_name='Waitlist Entry'),
        ),
    ]



# Generated by Django 4.2.19 on 2025-04-01 04:38

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0014_alter_activefleet_fleet'),
    ]

    operations = [
        migrations.AlterField(
            model_name='activefleet',
            name='fleet',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.fleet', verbose_name='Fleet'),
        ),
    ]



from corptools.models import EveItemType
from ninja import NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema, HullSchema
from incursions.models.waitlist import FittingHistory, FleetActivity

logger = get_extension_logger(__name__)
api = NinjaAPI()


class ActivityEntrySchema(Schema):
    ship: HullSchema
    logged_at: int
    time_in_fleet: int


class ActivitySummaryEntrySchema(Schema):
    ship: HullSchema
    time_in_fleet: int


class ActivityResponse(Schema):
    activity: list[ActivityEntrySchema]
    summary: list[ActivitySummaryEntrySchema]


class FleetCompEntrySchema(Schema):
    ship: HullSchema
    character: CharacterSchema
    logged_at: int
    time_in_fleet: int
    is_boss: bool


class HistoryFleetCompResponse(Schema):
    fleets: dict[int, list[FleetCompEntrySchema]]


class XupHistoryLineSchema(Schema):
    logged_at: int
    dna: str
    implants: list[int]
    ship: HullSchema


class XupHistorySchema(Schema):
    xups: list[XupHistoryLineSchema]


class SkillHistoryResponseLineSchema(Schema):
    skill_id: int
    old_level: int
    new_level: int
    logged_at: int
    name: str


class SkillHistoryResponse(Schema):
    history: list[SkillHistoryResponseLineSchema]
    ids: dict[str, int]


def setup(api: NinjaAPI) -> None:
    HistoryAPIEndpoints(api)


class HistoryAPIEndpoints:

    tags = ["History"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/history/fleet", response=ActivityResponse, tags=self.tags)
        def fleet_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                logger.warning(f"User {request.user} denied access to fleet history for character {character_id}")
                return 403, {"error": "Permission denied"}

            activities = FleetActivity.objects.filter(character_id=character_id).select_related("ship").order_by("-first_seen")
            time_by_ship: dict[int, int] = {}
            activity_entries: list[ActivityEntrySchema] = []

            for act in activities:
                time_in_fleet = act.last_seen - act.first_seen
                ship_id = act.ship.type_id
                ship_name = act.ship.name
                time_by_ship[ship_id] = time_by_ship.get(ship_id, 0) + time_in_fleet

                activity_entries.append(ActivityEntrySchema(
                    ship=HullSchema(id=ship_id, name=ship_name),
                    logged_at=act.first_seen,
                    time_in_fleet=time_in_fleet,
                ))

            summary_entries: list[ActivitySummaryEntrySchema] = [
                ActivitySummaryEntrySchema(ship=HullSchema(id=ship_id, name=EveItemType.objects.only("name").get(pk=ship_id).name), time_in_fleet=total_time)
                for ship_id, total_time in time_by_ship.items()
            ]
            summary_entries.sort(key=lambda s: s.time_in_fleet, reverse=True)

            logger.info(f"Fleet history returned for character {character_id} with {len(activity_entries)} entries")
            return {"activity": activity_entries, "summary": summary_entries}

        @api.get("/history/fleet-comp", response=HistoryFleetCompResponse, tags=self.tags)
        def fleet_comp(request, time: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                logger.warning(f"User {request.user} denied access to fleet composition at time {time}")
                return 403, {"error": "Permission denied"}

            comp_entries = FleetActivity.objects.select_related("character", "ship", "fleet", "fleet__boss").filter(first_seen__lte=time, last_seen__gte=time)
            fleets: dict[int, list[FleetCompEntrySchema]] = {}

            for entry in comp_entries:
                fleet_id = entry.fleet.pk
                is_boss = entry.fleet.boss_id == entry.character.pk
                fleets.setdefault(fleet_id, []).append(FleetCompEntrySchema(
                    ship=HullSchema.from_orm(entry.ship),
                    character=CharacterSchema.from_orm(entry.character),
                    logged_at=entry.first_seen,
                    time_in_fleet=entry.last_seen - entry.first_seen,
                    is_boss=is_boss,
                ))

            logger.info(f"Fleet composition at {time} returned with {sum(len(v) for v in fleets.values())} members")
            return {"fleets": fleets}

        @api.get("/history/xup", response=XupHistorySchema, tags=self.tags)
        def xup_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                logger.warning(f"User {request.user} denied access to XUP history for character {character_id}")
                return 403, {"error": "Permission denied"}

            xup_lines: list[XupHistoryLineSchema] = []
            fittings = FittingHistory.objects.filter(character_id=character_id).select_related("fit__ship", "implant_set").order_by("-pk")

            for xup in fittings:
                try:
                    implants_list = [int(i) for i in xup.implant_set.implants.split(':') if i]
                except Exception:
                    implants_list = []

                ship_schema = HullSchema().from_orm(xup.fit.ship)
                xup_lines.append(XupHistoryLineSchema(logged_at=xup.logged_at, dna=xup.fit.dna, implants=implants_list, ship=ship_schema))

            logger.info(f"XUP history returned for character {character_id} with {len(xup_lines)} entries")
            return {"xups": xup_lines}



from __future__ import annotations

from datetime import datetime

from ninja import NinjaAPI, Schema

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema
from incursions.models.waitlist import Badge, CharacterBadges


class BadgeSchema(Schema):
    id: int
    name: str
    letter: str
    type: str
    color: str
    order: int
    member_count: int
    exclude_badge_id: int | None


class CharacterBadgesSchema(Schema):
    badge: BadgeSchema
    character: CharacterSchema
    granted_by: CharacterSchema | None
    granted_at: datetime


class AssignBadgeRequest(Schema):
    character_id: int


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    BadgesAPIEndpoints(api)


class BadgesAPIEndpoints:

    tags = ["Badges"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/badges", response={200: list[BadgeSchema], 403: dict}, tags=["Badges"])
        def list_badges(request) -> tuple[int, list[BadgeSchema] | dict]:
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning(f"Permission denied to list badges for user {request.user}")
                return 403, {"error": "Permission denied"}

            badges = Badge.objects.all()
            result = [BadgeSchema.from_orm(b) for b in badges]
            logger.info(f"Returning {len(result)} badges to user {request.user}")
            return 200, result

        @api.get("/badges/{badge_id}/members", response={200: list[CharacterBadgesSchema], 403: dict, 404: dict}, tags=self.tags)
        def get_badge_members(request, badge_id: int) -> tuple[int, list[CharacterBadgesSchema] | dict]:
            if not (request.user.has_perm("incursions.waitlist_badges_view") or request.user.has_perm("incursions.waitlist_badges_manage")):
                logger.warning(f"Permission denied to view badge members for user {request.user}")
                return 403, {"error": "Permission denied"}

            badge = Badge.objects.filter(id=badge_id).first()
            if not badge:
                logger.error(f"Badge {badge_id} not found")
                return 404, {"error": "Badge not found"}

            characters = CharacterBadges.objects.filter(badge=badge).select_related("character", "granted_by", "badge")
            logger.info(f"Returning {characters.count()} badge members for badge {badge_id} to user {request.user}")
            return 200, [CharacterBadgesSchema.from_orm(cb) for cb in characters]

        @api.post("/badges/{badge_id}/members", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def assign_badge(request, badge_id: int, payload: AssignBadgeRequest) -> tuple[int, dict]:
            if not request.user.has_perm("incursions.waitlist_badges_manage"):
                logger.warning(f"Permission denied to assign badge by user {request.user}")
                return 403, {"error": "Permission denied"}

            character = EveCharacter.objects.filter(character_id=payload.character_id).first()
            if not character:
                logger.error(f"Character {payload.character_id} not found")
                return 404, {"error": "EveCharacter not found"}

            badge = Badge.objects.filter(id=badge_id).first()
            if not badge:
                logger.error(f"Badge {badge_id} not found")
                return 404, {"error": "Badge not found"}

            if CharacterBadges.objects.filter(character=character, badge=badge).exists():
                logger.warning(f"Character {payload.character_id} already has badge {badge_id}")
                return 400, {"error": "EveCharacter already has this badge"}

            CharacterBadges.objects.create(badge=badge, character=character, granted_by=get_main_character_from_user(request.user))
            logger.info(f"Assigned badge {badge_id} to character {payload.character_id} by user {request.user}")
            return 200, {"status": "Badge assigned"}

        @api.delete("/badges/{badge_id}/members/{character_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def revoke_badge(request, badge_id: int, character_id: int) -> tuple[int, dict]:
            if not request.user.has_perm("incursions.waitlist_badges_manage"):
                logger.warning(f"Permission denied to revoke badge by user {request.user}")
                return 403, {"error": "Permission denied"}

            character = EveCharacter.objects.filter(character_id=character_id).first()
            if not character:
                logger.error(f"Character {character_id} not found")
                return 404, {"error": "Character not found"}

            badge = Badge.objects.filter(id=badge_id).first()
            if not badge:
                logger.error(f"Badge {badge_id} not found")
                return 404, {"error": "Badge not found"}

            badge_qs = CharacterBadges.objects.filter(character=character, badge=badge)
            if not badge_qs.exists():
                logger.warning(f"Badge assignment for character {character_id} and badge {badge_id} not found")
                return 404, {"error": "Badge assignment not found"}

            deleted, _ = badge_qs.delete()
            logger.info(f"Revoked badge {badge_id} from character {character_id} by user {request.user}")
            return 200, {"status": "Badge revoked"}



# Generated by Django 4.2.19 on 2025-04-02 01:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0016_incursionsconfig_waitlist_webhooks_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='incursionsconfig',
            name='status_webhooks',
            field=models.ManyToManyField(blank=True, related_name='+', to='incursions.webhook', verbose_name='Destination Webhook for Incursion Updates'),
        ),
        migrations.AlterField(
            model_name='incursionsconfig',
            name='waitlist_webhooks',
            field=models.ManyToManyField(blank=True, related_name='+', to='incursions.webhook', verbose_name='Destination Webhook for Waitlist Updates'),
        ),
    ]



# Generated by Django 4.2.19 on 2025-03-31 13:14

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0118_alter_skill_id'),
        ('incursions', '0012_rename_hull_approvedskills_ship_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='activefleet',
            options={'default_permissions': (), 'verbose_name': 'Waitlist - Fleet (Active)'},
        ),
        migrations.AlterField(
            model_name='activefleet',
            name='fleet',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.fleet', verbose_name='Fleet'),
        ),
        migrations.AlterField(
            model_name='approvedskills',
            name='ship',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Ship'),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import redirect, render

from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

from incursions import __version__
from incursions.app_settings import (
    INCURSIONS_SCOPES_BASE, INCURSIONS_SCOPES_FC,
)

logger = get_extension_logger(__name__)


@login_required
def waitlist(request) -> HttpResponse:
    return render(request, 'incursions/base-bs5.html', context={"version": __version__, "app_name": "incursions/waitlist", "page_title": "Waitlist"})


# @login_required
# @token_required(scopes=INCURSIONS_SCOPES_BASE)
# def add_char_base(request, token):

#     return redirect('incursions:waitlist')

@login_required
@token_required(scopes=INCURSIONS_SCOPES_BASE + INCURSIONS_SCOPES_FC)
def add_char_fc(request, token):

    return redirect('incursions:waitlist')



# Generated by Django 4.2 on 2025-04-10 05:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0027_alter_waitlistentryfit_tags_delete_skillhistory'),
    ]

    operations = [
        migrations.AlterField(
            model_name='badge',
            name='color',
            field=models.CharField(choices=[('red', 'Red'), ('green', 'Green'), ('blue', 'Blue'), ('yellow', 'Yellow'), ('purple', 'Purple'), ('neutral', 'Neutral'), ('cyan', 'Cyan')], default='neutral', max_length=16, verbose_name='Badge Color'),
        ),
        migrations.AlterField(
            model_name='badge',
            name='type',
            field=models.CharField(choices=[('shield', 'Shield')], default='shield', max_length=16, verbose_name='Badge Type'),
        ),
    ]



import json
from datetime import datetime

from corptools.models import EveItemType
from ninja import NinjaAPI, Schema

from django.db import transaction
from django.db.models import Count, Prefetch
from django.http import HttpRequest
from django.shortcuts import get_object_or_404
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.evecharacter import (
    get_main_character_from_evecharacter,
)
from allianceauth.framework.api.user import (
    get_all_characters_from_user, get_main_character_from_user,
)
from allianceauth.services.hooks import get_extension_logger

from incursions.api.badges import BadgeSchema
from incursions.api.helpers.fittings import FittingParser
from incursions.api.helpers.sse import SSEClient, SSEEvent
from incursions.api.schema import CharacterSchema, HullSchema
from incursions.app_settings import SSE_SECRET, SSE_SITE_URL
from incursions.models.waitlist import (
    ActiveFleet, ApprovedFitting, CharacterBadges, Fitting, FleetSquad,
    ImplantSet, Waitlist, WaitlistCategory, WaitlistCategoryRule,
    WaitlistEntry, WaitlistEntryFit,
)
from incursions.providers import get_character_implants, invite_to_fleet

logger = get_extension_logger(__name__)
sse_client = SSEClient(SSE_SITE_URL, SSE_SECRET)


# Schemas
class WaitlistEntryFitSchema(Schema):
    id: int
    approved: bool
    category: str
    dna: str
    ship: HullSchema
    character: CharacterSchema
    hours_in_fleet: int
    review_comment: str | None = None
    implants: list[int]
    fit_analysis: dict | None = None
    is_alt: bool
    messagexup: str | None = None
    badges: list[BadgeSchema]


class WaitlistEntrySchema(Schema):
    id: int
    main_character: CharacterSchema | None = None
    joined_at: datetime
    can_remove: bool
    fits: list[WaitlistEntryFitSchema]


class WaitlistResponse(Schema):
    open: bool
    waitlist: list[WaitlistEntrySchema] | None = None
    categories: list[str]


class XupRequest(Schema):
    waitlist_id: int
    character_id: int
    eft: str
    is_alt: bool
    messagexup: str


class RemoveWaitlistEntryRequest(Schema):
    id: int


class RemoveWaitlistEntryFitRequest(Schema):
    id: int


class InviteRequest(Schema):
    id: int


class ApproveRequest(Schema):
    id: int


class RejectRequest(Schema):
    id: int
    review_comment: str


class SetOpenRequest(Schema):
    open: bool
    waitlist_id: int


class EmptyWaitlistRequest(Schema):
    waitlist_id: int


class WaitlistUpdateSchema(Schema):
    waitlist_id: int


api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    WaitlistAPIEndpoints(api)


class WaitlistAPIEndpoints:
    tags = ["Waitlist"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/waitlist", response={200: WaitlistResponse, 403: dict, 404: dict}, tags=self.tags)
        def get_waitlist(request: HttpRequest) -> WaitlistResponse | tuple[int, dict]:
            logger.info("Fetching waitlist for user %s", request.user)
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning("User %s does not have permission incursions.basic_waitlist", request.user)
                return 403, {"error": "Permission denied"}

            waitlist = Waitlist.get_solo()
            categories: list[str] = list(WaitlistCategory.objects.values_list("name", flat=True))
            request_main_character = get_main_character_from_user(request.user)

            context_manager: bool = request.user.has_perm("incursions.waitlist_manage_waitlist")
            context_wl_approver: bool = context_manager or request.user.has_perm("incursions.waitlist_manage_waitlist_approve_fits")
            context_implants: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_implants_view")
            context_stats: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_stats_view")
            context_quantity: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_a")
            context_ship: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_b")
            context_time: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_c")
            context_name: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_d")

            if not waitlist.is_open:
                logger.info("Waitlist is closed.")
                return WaitlistResponse(open=False, waitlist=None, categories=categories)

            waitlist_entries = WaitlistEntry.objects.filter(waitlist=waitlist).select_related("main_character", "waitlist")

            response_waitlist: list[WaitlistEntrySchema] = []
            for waitlist_entry in waitlist_entries:
                fits_data: list[WaitlistEntryFitSchema] = []
                is_self: bool = request_main_character == waitlist_entry.main_character
                for wl_fit in WaitlistEntryFit.objects.filter(
                    waitlist_entry=waitlist_entry
                ).select_related(
                    "implant_set", "fit", "category", "character"
                ).prefetch_related(
                    Prefetch(
                        "character__incursions_badge",
                        queryset=CharacterBadges.objects.select_related("badge"),
                        to_attr="prefetched_badges"
                    )
                ):

                    if wl_fit.implant_set and wl_fit.implant_set.implants:
                        implants_list = json.loads(wl_fit.implant_set.implants)
                    else:
                        implants_list: list[str] = []

                    fits_data.append(
                        WaitlistEntryFitSchema(
                            id=wl_fit.pk,
                            approved=wl_fit.approved,
                            category=wl_fit.category.name,
                            ship=HullSchema(
                                id=wl_fit.fit.ship.type_id if (is_self or context_wl_approver or context_ship) else 670,  # Capsule for memes (and a real icon)
                                name=wl_fit.fit.ship.name if (is_self or context_wl_approver or context_ship) else "Hidden",
                            ),
                            character=CharacterSchema(
                                character_name=wl_fit.character.character_name if (is_self or context_name) else "Hidden",
                                character_id=wl_fit.character.character_id if (is_self or context_name) else 0,
                                corporation_id=wl_fit.character.corporation_id if (is_self or context_name) else 0,
                            ),
                            hours_in_fleet=wl_fit.cached_time_in_fleet // 3600 if context_stats else 0,
                            review_comment=wl_fit.review_comment if (is_self or context_wl_approver) else None,
                            dna=wl_fit.fit.dna if (is_self or context_wl_approver) else "",
                            implants=implants_list if (is_self or context_wl_approver or context_implants) else [],
                            fit_analysis=wl_fit.fit_analysis if (is_self or context_wl_approver or context_implants) else None,
                            is_alt=wl_fit.is_alt if (is_self or context_wl_approver or context_quantity) else False,
                            messagexup=wl_fit.messagexup if (is_self or context_wl_approver) else None,
                            badges=[BadgeSchema.from_orm(cb.badge) for cb in getattr(wl_fit.character, "prefetched_badges", [])]
                        )
                    )
                response_waitlist.append(
                    WaitlistEntrySchema(
                        id=waitlist_entry.pk,
                        main_character=CharacterSchema(
                            character_name=waitlist_entry.main_character.character_name if (is_self or context_name) else "Hidden",
                            character_id=waitlist_entry.main_character.character_id if (is_self or context_name) else 0,
                        ),
                        joined_at=waitlist_entry.joined_at if (is_self or context_time) else now(),
                        can_remove=True if (is_self or context_wl_approver) else False,
                        fits=fits_data,
                    )
                )
            logger.info("Returning waitlist with %d entries", len(response_waitlist))
            return WaitlistResponse(open=True, waitlist=response_waitlist, categories=categories)

        @api.post("/waitlist/xup", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def xup(request: HttpRequest, payload: XupRequest) -> tuple[int, dict]:
            logger.info(
                "Processing xup for waitlist_id: %s and character_id: %s",
                payload.waitlist_id,
                payload.character_id,
            )
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning("User %s lacks basic_waitlist permission", request.user)
                return 403, {"error": "Permission denied"}

            waitlist = Waitlist.get_solo()
            character = get_object_or_404(EveCharacter, character_id=payload.character_id)
            fitting_parser = FittingParser.from_eft(payload.eft)

            fitting, _ = Fitting.objects.get_or_create(
                dna=fitting_parser.to_dna(),
                ship=EveItemType.objects.get_or_create_from_esi(fitting_parser.ship)[0],
            )
            implant_set, _ = ImplantSet.objects.get_or_create(implants=list(get_character_implants(character.character_id)[0]))
            try:
                rule = WaitlistCategoryRule.objects.select_related("waitlist_category").get(ship=fitting.ship)
                waitlist_category = rule.waitlist_category
            except WaitlistCategoryRule.DoesNotExist:
                waitlist_category, _ = WaitlistCategory.objects.get_or_create(name="Sponge")

            try:
                fit_analysis = fitting_parser.to_analysis(doctrine=FittingParser.from_dna(ApprovedFitting.objects.get(ship=fitting.ship).dna))
                # Ariel Badges Add Bad Fit?
            except ApprovedFitting.DoesNotExist:
                fit_analysis = {}
            with transaction.atomic():
                waitlist_entry, _ = WaitlistEntry.objects.get_or_create(main_character=get_main_character_from_evecharacter(character), waitlist=waitlist)
                new_fit, created = WaitlistEntryFit.objects.update_or_create(
                    character=character,
                    waitlist_entry=waitlist_entry,
                    defaults={
                        # "approved": False,
                        "category": waitlist_category,
                        "cached_time_in_fleet": 0,
                        "is_alt": payload.is_alt,  # ARIEL i can probably calculate this on my end, not frontend
                        "messagexup": payload.messagexup,
                        "tags": "",  # ARIEL i can probably calculate this on my end, not frontend
                        "fit": fitting,
                        "implant_set": implant_set,
                        "fit_analysis": fit_analysis if fit_analysis else {},
                    }
                )
            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit([SSEEvent.new_json(
                topic="waitlist", event="waitlist_update", data=sse_payload
            )])
            boss_char = get_main_character_from_evecharacter(ActiveFleet.get_solo().fleet.boss)

            if created is True:
                sse_client.submit([SSEEvent.new(
                    topic=f"account;{boss_char.character_id}",
                    event="message",
                    data="New x-up in waitlist",
                )])
                logger.info("X-up recorded with fit_id %s", new_fit.pk)
                return 200, {"status": "X-up recorded", "fit_id": new_fit.pk}
            else:
                sse_client.submit([SSEEvent.new(
                    topic=f"account;{boss_char.character_id}",
                    event="message",
                    data="Updated x-up in waitlist",
                )])
                logger.info("X-up updated with fit_id %s", new_fit.pk)
                return 200, {"status": "X-up updated", "fit_id": new_fit.pk}

        @api.post("/waitlist/remove_waitlistentry", response={200: dict, 403: dict}, tags=self.tags)
        def remove_waitlistentry(request: HttpRequest, payload: RemoveWaitlistEntryRequest) -> tuple[int, dict]:
            # Called by Self, this is the big ol remove entry button
            # Drops the whole WL Entry, + all associated fits
            logger.info("Removing waitlist entry x with id %s", payload.id)

            waitlist = Waitlist.get_solo()

            waitlist_entry = WaitlistEntry.objects.filter(pk=payload.id).first()
            if not waitlist_entry:
                logger.warning("Waitlist entry %s does not exist", payload.id)
                return 404, {"error": "Waitlist entry not found"}

            if not (
                waitlist_entry.main_character == get_main_character_from_user(request.user)
                or request.user.has_perm("incursions.waitlist_manage_waitlist")
            ):
                logger.warning("User %s not authorized to remove waitlist entry %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}

            with transaction.atomic():
                waitlist_entry.delete()

            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Waitlist entry %s removed", payload.id)
            return 200, {"status": "Waitlist X-Up removed"}

        @api.post("/waitlist/remove_waitlistentryfit", response={200: dict, 403: dict}, tags=self.tags)
        def remove_waitlistentryfit(request: HttpRequest, payload: RemoveWaitlistEntryFitRequest) -> tuple[int, dict]:
            # This means an FC is removing your fit, not usually called by self
            logger.info("Removing fit with id %s", payload.id)
            entryfit = get_object_or_404(WaitlistEntryFit, pk=payload.id)
            waitlist = Waitlist.get_solo()

            if not (entryfit.character in get_all_characters_from_user(request.user) or request.user.has_perm("incursions.waitlist_manage_waitlist")):
                logger.warning("User %s not authorized to remove fit %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}

            with transaction.atomic():
                WaitlistEntryFit.objects.filter(pk=payload.id).delete()
                WaitlistEntry.objects.annotate(fit_count=Count('waitlistentryfit')).filter(fit_count=0).delete()

            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit([SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)])
            logger.info("Waitlist fit %s removed", payload.id)
            return 200, {"status": "Waitlist X-Up Fit removed"}

        @api.post("/waitlist/invite", response={200: dict, 403: dict}, tags=self.tags)
        def invite(request: HttpRequest, payload: InviteRequest) -> tuple[int, dict]:
            logger.info(f"Inviting waitlist entry fit id:{payload.id}", )
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                logger.warning("User %s not authorized to invite", request.user)
                return 403, {"error": "Permission denied"}
            active_fleet = ActiveFleet.get_solo()
            waitlist_entry_fit = get_object_or_404(WaitlistEntryFit, pk=payload.id)
            fleet_squad = get_object_or_404(FleetSquad, fleet=active_fleet.fleet, category__name=waitlist_entry_fit.category.name)
            squad_id = fleet_squad.squad_id
            wing_id = fleet_squad.wing_id
            logger.info(f"Inviting character:{waitlist_entry_fit.character.character_id} for waitlist entry fit id:{payload.id}")

            invite_to_fleet(
                boss_character_id=active_fleet.fleet.boss.character_id,
                fleet_id=active_fleet.fleet.pk,
                character_id=waitlist_entry_fit.character.character_id,
                squad_id=squad_id if squad_id else None,
                wing_id=wing_id if wing_id else None,
                role="squad_member",
            )
            sse_client.submit([SSEEvent.new(
                topic=f"account;{get_main_character_from_evecharacter(waitlist_entry_fit.character).character_id}",
                event="wakeup",
                data=f"{active_fleet.fleet.boss.character_name} has invited your {waitlist_entry_fit.fit.ship.name} to fleet.",
            )])
            logger.info(f"Character:{waitlist_entry_fit.character.character_id} invited for waitlist entry:{payload.id}")
            return 200, {"status": f"Character {waitlist_entry_fit.character.character_id} invited to waitlist entry {payload.id}"}

        @api.post("/waitlist/approve", tags=self.tags)
        def approve_fit(request: HttpRequest, payload: ApproveRequest) -> dict:
            logger.info("Approving fit with id %s", payload.id)
            if not (
                request.user.has_perm("incursions.waitlist_manage_waitlist")
                or request.user.has_perm("incursions.waitlist_manage_waitlist_approve_fits")
            ):
                logger.warning("User %s not authorized to approve fit %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                fit = get_object_or_404(WaitlistEntryFit, pk=payload.id)
                fit.approved = True
                fit.save(update_fields=["approved"])
            sse_payload = WaitlistUpdateSchema(waitlist_id=fit.waitlist_entry.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Fit %s approved", payload.id)
            return {"status": f"Fit {fit.pk} approved"}

        @api.post("/waitlist/reject", tags=self.tags)
        def reject_fit(request: HttpRequest, payload: RejectRequest) -> tuple[int, dict]:
            logger.info("Rejecting fit with id %s", payload.id)
            if not (
                request.user.has_perm("incursions.waitlist_manage_waitlist")
                or request.user.has_perm("incursions.waitlist_manage_waitlist_approve_fits")
            ):
                logger.warning("User %s not authorized to reject fit %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                fit = get_object_or_404(WaitlistEntryFit, pk=payload.id)
                fit.approved = False
                fit.review_comment = payload.review_comment
                fit.save(update_fields=["approved", "review_comment"])
            sse_payload = WaitlistUpdateSchema(waitlist_id=fit.waitlist_entry.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Fit %s rejected", payload.id)
            return 200, {"status": f"Fit {fit.pk} rejected"}

        @api.post("/waitlist/set_open", tags=self.tags)
        def set_open(request: HttpRequest, payload: SetOpenRequest) -> tuple[int, str | dict]:
            logger.info("Setting waitlist %s open state to %s", payload.waitlist_id, payload.open)
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                logger.warning("User %s not authorized to set open state", request.user)
                return 403, {"error": "Permission denied"}

            waitlist = Waitlist.get_solo()
            waitlist.is_open = payload.open
            waitlist.save(update_fields=["is_open"])
            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="open", data=sse_payload)]
            )
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Waitlist %s open state set to %s", payload.waitlist_id, payload.open)
            return 200, f"Waitlist is now {'Open' if payload.open is True else 'Closed'}."

        @api.post("/waitlist/empty", tags=self.tags)
        def empty_waitlist(request: HttpRequest, payload: EmptyWaitlistRequest) -> dict:
            logger.info("Emptying waitlist with id %s", payload.waitlist_id)
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                logger.warning("User %s not authorized to empty waitlist", request.user)
                return 403, {"error": "Permission denied"}

            waitlist = Waitlist.get_solo()

            with transaction.atomic():
                WaitlistEntryFit.objects.filter(waitlist_entry__waitlist=waitlist).delete()
                WaitlistEntry.objects.filter(waitlist=waitlist).delete()
                waitlist.is_open = False
                waitlist.save()

            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Waitlist %s emptied", payload.waitlist_id)
            return 200, {"status": "OK"}



from django.db import migrations


def migrate_incursion_locations(apps, schema_editor):
    Incursion = apps.get_model('incursions', 'Incursion')
    MapConstellation = apps.get_model('corptools', 'MapConstellation')
    MapSystem = apps.get_model('corptools', 'MapSystem')

    for incursion in Incursion.objects.all():
        if incursion.constellation:
            try:
                incursion.migration_constellation = MapConstellation.objects.get(constellation_id=incursion.constellation.constellation_id)
            except MapConstellation.DoesNotExist:
                pass

        if incursion.staging_solar_system:
            try:
                incursion.migration_staging_solar_system = MapSystem.objects.get(
                    system_id=incursion.staging_solar_system.solar_system_id
                )
            except MapSystem.DoesNotExist:
                pass

        incursion.save()

        system_ids = incursion.infested_solar_systems.values_list('solar_system_id', flat=True)
        mapped_systems = MapSystem.objects.filter(system_id__in=system_ids)
        incursion.migration_infested_solar_systems.set(mapped_systems)


def reverse_migration(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0118_alter_skill_id'),
        ('incursions', '0020_alter_incursion_migration_infested_solar_systems'),
    ]

    operations = [
        migrations.RunPython(migrate_incursion_locations, reverse_migration),
    ]



from corptools.models import CharacterAudit, EveItemGroup, EveItemType, Skill
from ninja import NinjaAPI, Schema

from django.http import HttpRequest
from django.shortcuts import get_object_or_404

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_all_characters_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.app_settings import EVECATEGORY_SKILLS
from incursions.models.waitlist import ApprovedSkills

SKILL_GROUP_GUNNERY = 255
SKILL_GROUP_MISSILES = 256
SKILL_GROUP_SPACESHIP_COMMAND = 257
SKILL_GROUP_ARMOR = 1210
SKILL_GROUP_RIGGING = 269
SKILL_GROUP_SHIELDS = 1209
SKILL_GROUP_DRONES = 273
SKILL_GROUP_NAVIGATION = 275
SKILL_GROUP_NEURAL_ENHANCEMENT = 1220
SKILL_GROUP_ENGINEERING = 1216
SKILL_GROUP_ELECTRONIC_SYSTEMS = 272
SKILL_GROUP_FLEET_SUPPORT = 258
SKILL_GROUP_SUBSYSTEMS = 1240
SKILL_GROUP_TARGETING = 1213

RELEVANT_SKILL_GROUPS = [
    SKILL_GROUP_ARMOR,
    SKILL_GROUP_DRONES,
    SKILL_GROUP_ELECTRONIC_SYSTEMS,
    SKILL_GROUP_ENGINEERING,
    SKILL_GROUP_FLEET_SUPPORT,
    SKILL_GROUP_GUNNERY,
    SKILL_GROUP_MISSILES,
    SKILL_GROUP_NAVIGATION,
    SKILL_GROUP_NEURAL_ENHANCEMENT,
    SKILL_GROUP_RIGGING,
    SKILL_GROUP_SHIELDS,
    SKILL_GROUP_SPACESHIP_COMMAND,
    SKILL_GROUP_SUBSYSTEMS,
    SKILL_GROUP_TARGETING,
]


class SkillTierSchema(Schema):
    min: int | None
    elite: int | None
    gold: int | None


class SkillsResponse(Schema):
    current: dict[int, int]
    ids: dict[str, int]
    categories: dict[str, list[int]]
    requirements: dict[str, dict[int, SkillTierSchema]]


logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    SkillsAPIEndpoints(api)


class SkillsAPIEndpoints:

    tags = ["Skills"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/skills/{character_id}", response={200: SkillsResponse, 403: dict}, tags=self.tags)
        def list_skills(request: HttpRequest, character_id: int):
            character = get_object_or_404(EveCharacter.objects.only("pk"), character_id=character_id)

            if not (character in get_all_characters_from_user(request.user) or request.user.has_perm("incursions.waitlist_skills_view")):
                logger.warning(f"User {request.user} denied access to skills for character {character_id}")
                return 403, {"error": "Permission denied"}

            character_audit = get_object_or_404(CharacterAudit.objects.select_related("character"), character=character)
            skills = Skill.objects.filter(character=character_audit, skill_name__group_id__in=RELEVANT_SKILL_GROUPS).select_related("skill_name")

            requirements = {
                shipskills.ship.name: {
                    x.skill.type_id: SkillTierSchema.from_orm(x)
                    for x in shipskills.skill_checks.all()
                }
                for shipskills in ApprovedSkills.objects.select_related("ship").prefetch_related("skill_checks__skill")
            }

            current_skills = {skill.skill_name.type_id: skill.trained_skill_level for skill in skills}
            ids = {skill.name: skill.type_id for skill in EveItemType.objects.filter(group__category_id=EVECATEGORY_SKILLS).only("name", "type_id")}
            categories = {
                group.name: list(EveItemType.objects.filter(group=group).values_list("type_id", flat=True))
                for group in EveItemGroup.objects.filter(group_id__in=RELEVANT_SKILL_GROUPS).only("name")
            }

            logger.info(f"User {request.user} retrieved skills for character {character_id}")
            return SkillsResponse(current=current_skills, ids=ids, categories=categories, requirements=requirements)



from __future__ import annotations

import json
from datetime import datetime

from ninja import NinjaAPI, Schema

from django.utils.timezone import now

from allianceauth.framework.tests.test_api_user import (
    get_main_character_from_user,
)
from allianceauth.services.hooks import get_extension_logger

from incursions.api.helpers.sse import SSEClient, SSEEvent
from incursions.api.schema import CharacterSchema
from incursions.app_settings import SSE_SECRET, SSE_SITE_URL
from incursions.models.waitlist import Announcement


class RequestPayloadSchema(Schema):
    message: str
    is_alert: bool
    pages: str | None


class AnnouncementSchema(Schema):
    id: int
    message: str
    is_alert: bool
    pages: str | None
    created_by: CharacterSchema | None
    created_at: datetime
    revoked_by: CharacterSchema | None
    revoked_at: datetime | None


sse_client = SSEClient(SSE_SITE_URL, SSE_SECRET)
logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    AnnouncementAPIEndpoints(api)


class AnnouncementAPIEndpoints:

    tags = ["Announcements"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/announcements", response={200: list[AnnouncementSchema], 403: dict}, tags=self.tags)
        def list_announcements(request) -> tuple[int, list[AnnouncementSchema] | dict]:
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning(f"Permission denied for listing announcements by user {request.user}")
                return 403, {"error": "Permission denied"}

            announcements = list(Announcement.objects.filter(revoked_at__isnull=True).select_related("created_by", "revoked_by"))
            logger.info(f"Returning {len(announcements)} active announcements to user {request.user}")
            return 200, announcements

        @api.post("/announcements", tags=self.tags, response={200: dict, 403: dict})
        def create_announcement(request, payload: RequestPayloadSchema) -> tuple[int, dict]:
            if not request.user.has_perm("incursions.waitlist_announcements_manage"):
                logger.warning(f"Permission denied for creating announcement by user {request.user}")
                return 403, {"error": "Permission denied"}

            announcement = Announcement.objects.create(
                message=payload.message,
                is_alert=payload.is_alert,
                pages=payload.pages,
                created_by=get_main_character_from_user(request.user),
                # created_at=now(), # auto_now_add
            )
            logger.info(f"Created announcement #{announcement.pk} by user {request.user}")
            sse_client.submit([SSEEvent.new_json(topic="announcements", event="announcement;new", data=json.loads(AnnouncementSchema.from_orm(announcement).model_dump_json()))])
            return 200, {"id": announcement.pk, "status": "Created"}

        @api.put("/announcements/{announcement_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def update_announcement(request, announcement_id: int, payload: RequestPayloadSchema) -> tuple[int, dict]:
            if not request.user.has_perm("incursions.waitlist_announcements_manage"):
                logger.warning(f"Permission denied for updating announcement by user {request.user}")
                return 403, {"error": "Permission denied"}

            announcement = Announcement.objects.get(id=announcement_id, revoked_at__isnull=True)
            if not announcement:
                logger.error(f"Announcement #{announcement_id} not found for update")
                return 404, {"error": "Announcement not found"}

            announcement.message = payload.message
            announcement.is_alert = payload.is_alert
            announcement.pages = payload.pages
            announcement.created_by = get_main_character_from_user(request.user)
            announcement.save(update_fields=["message", "is_alert", "pages", "created_by"])

            logger.info(f"Updated announcement #{announcement.pk} by user {request.user}")
            sse_client.submit([SSEEvent.new_json(topic="announcements", event="announcement;updated", data=json.loads(AnnouncementSchema.from_orm(announcement).model_dump_json()))])
            return 200, {"status": "Updated"}

        @api.delete("/announcements/{announcement_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def revoke_announcement(request, announcement_id: int) -> tuple[int, dict]:
            if not request.user.has_perm("incursions.waitlist_announcements_manage"):
                logger.warning(f"Permission denied for revoking announcement by user {request.user}")
                return 403, {"error": "Permission denied"}

            announcement = Announcement.objects.get(id=announcement_id, revoked_at__isnull=True)
            if not announcement:
                logger.error(f"Announcement #{announcement_id} not found or already revoked")
                return 404, {"error": "Announcement not found or already revoked"}

            announcement.revoked_by = get_main_character_from_user(request.user)
            announcement.revoked_at = now()
            announcement.save(update_fields=["revoked_by", "revoked_at"])

            logger.info(f"Revoked announcement #{announcement.pk} by user {request.user}")
            sse_client.submit([SSEEvent.new_json(topic="announcements", event="announcement;updated", data=json.loads(AnnouncementSchema.from_orm(announcement).model_dump_json()))])
            return 200, {"status": "Revoked"}



# Generated by Django 4.2 on 2025-04-05 11:36

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0118_alter_skill_id'),
        ('incursions', '0018_fleet_closed_fleet_last_updated_fleet_open_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='incursion',
            name='migration_constellation',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.mapconstellation'),
        ),
        migrations.AddField(
            model_name='incursion',
            name='migration_infested_solar_systems',
            field=models.ManyToManyField(blank=True, null=True, related_name='+', to='corptools.mapsystem'),
        ),
        migrations.AddField(
            model_name='incursion',
            name='migration_staging_solar_system',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.mapsystem'),
        ),
    ]



# Generated by Django 4.2 on 2025-04-07 00:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0021_corptoolsmigration'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False, 'permissions': (('basic_waitlist', 'Waitlist Access'), ('waitlist_alts_view', 'View - all of a users Alts ! Danger?'), ('waitlist_announcements_manage', 'Announcements - Manage'), ('waitlist_badges_view', 'Badges - View'), ('waitlist_badges_manage', 'Badges - Manage'), ('waitlist_bans_view', 'Bans - View'), ('waitlist_bans_manage', 'Bans - Manage'), ('waitlist_documentation_view', 'Documentation - View'), ('waitlist_fleet_view', 'Fleet Composition'), ('waitlist_history_view', 'Fleet History'), ('waitlist_notes_view', 'Notes - View'), ('waitlist_notes_manage', 'Notes - Manage'), ('waitlist_search', 'Search'), ('waitlist_stats_view', 'Fleet Stats'), ('waitlist_skills_view', 'Skills'), ('waitlist_context_a', 'Context: Number of Pilots'), ('waitlist_context_b', 'Context: Ship Types'), ('waitlist_context_c', 'Context: Time in Waitlist'), ('waitlist_context_d', 'Context: Pilot Names'), ('waitlist_implants_view', 'Context: Implants'), ('waitlist_manage_waitlist', 'Waitlist Manage'), ('waitlist_manage_waitlist_approve_fits', 'Waitlist Manage Fits'), ('waitlist_esi_show_info', 'Show Info'), ('waitlist_esi_search', 'Search - ESI'))},
        ),
    ]
